import { BaseResponseDto } from '@app/common/dto/base.response.dto';
import { HttpException, HttpStatus, Injectable } from '@nestjs/common';
import {
  AddSubspaceDto,
  GetSpaceParam,
  GetSubSpaceParam,
  ModifySubspaceDto,
} from '../../dtos';
import { SuccessResponseDto } from '@app/common/dto/success.response.dto';
import {
  TypeORMCustomModel,
  TypeORMCustomModelFindAllQuery,
} from '@app/common/models/typeOrmCustom.model';
import { PageResponse } from '@app/common/dto/pagination.response.dto';
import { SubspaceDto } from '@app/common/modules/space/dtos';
import { In, QueryRunner } from 'typeorm';

import { SubspaceModelEntity } from '@app/common/modules/space-model';
import { ValidationService } from '../space-validation.service';
import { SubspaceRepository } from '@app/common/modules/space/repositories/subspace.repository';
import { ModifyAction } from '@app/common/constants/modify-action.enum';
import { SubspaceDeviceService } from './subspace-device.service';
import { ModifyTagDto } from 'src/space/dtos/tag/modify-tag.dto';
import { SpaceEntity } from '@app/common/modules/space/entities/space.entity';
import { SubspaceEntity } from '@app/common/modules/space/entities/subspace/subspace.entity';
import { ProcessTagDto } from 'src/tags/dtos';
import { TagService as NewTagService } from 'src/tags/services/tags.service';
import { SubspaceProductAllocationService } from './subspace-product-allocation.service';
import { SubspaceProductAllocationEntity } from '@app/common/modules/space/entities/subspace/subspace-product-allocation.entity';

@Injectable()
export class SubSpaceService {
  constructor(
    private readonly subspaceRepository: SubspaceRepository,
    private readonly validationService: ValidationService,
    private readonly newTagService: NewTagService,
    public readonly deviceService: SubspaceDeviceService,
    private readonly subspaceProductAllocationService: SubspaceProductAllocationService,
  ) {}

  async createSubspaces(
    subspaceData: Array<{
      subspaceName: string;
      space: SpaceEntity;
      subSpaceModel?: SubspaceModelEntity;
    }>,
    queryRunner: QueryRunner,
  ): Promise<SubspaceEntity[]> {
    try {
      const subspaceNames = subspaceData.map((data) => data.subspaceName);

      await this.checkExistingNamesInSpace(
        subspaceNames,
        subspaceData[0].space,
      );

      const subspaces = subspaceData.map((data) =>
        queryRunner.manager.create(this.subspaceRepository.target, data),
      );
      return await queryRunner.manager.save(subspaces);
    } catch (error) {
      throw new HttpException(
        `An unexpected error occurred while creating subspaces. ${error}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async createSubSpaceFromModel(
    subspaceModels: SubspaceModelEntity[],
    spaces: SpaceEntity[],
    queryRunner: QueryRunner,
  ): Promise<void> {
    if (!subspaceModels?.length || !spaces?.length) return;

    const subspaceData = [];

    for (const space of spaces) {
      for (const subSpaceModel of subspaceModels) {
        subspaceData.push({
          subspaceName: subSpaceModel.subspaceName,
          space,
          subSpaceModel,
        });
      }
    }
    await this.createSubspaces(subspaceData, queryRunner);
  }

  async createSubspacesFromDto(
    addSubspaceDtos: AddSubspaceDto[],
    space: SpaceEntity,
    queryRunner: QueryRunner,
    otherTags?: ProcessTagDto[],
    projectUuid?: string,
  ): Promise<SubspaceEntity[]> {
    try {
      this.checkForDuplicateNames(
        addSubspaceDtos.map(({ subspaceName }) => subspaceName),
      );

      const subspaceData = addSubspaceDtos.map((dto) => ({
        subspaceName: dto.subspaceName,
        space,
      }));

      const subspaces = await this.createSubspaces(subspaceData, queryRunner);
      await Promise.all(
        addSubspaceDtos.map(async (dto, index) => {
          const subspace = subspaces[index];

          const allTags = [...(dto.tags || []), ...(otherTags || [])];

          if (allTags.length) {
            const processedTags = await this.newTagService.processTags(
              allTags,
              projectUuid,
              queryRunner,
            );

            await this.subspaceProductAllocationService.createSubspaceProductAllocations(
              subspace,
              processedTags,
              queryRunner,
            );
          }
        }),
      );
      return subspaces;
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }

      throw new HttpException(
        'Failed to save subspaces due to an unexpected error.',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async createSubspace(
    addSubspaceDto: AddSubspaceDto,
    params: GetSpaceParam,
  ): Promise<BaseResponseDto> {
    const space =
      await this.validationService.validateSpaceWithinCommunityAndProject(
        params.communityUuid,
        params.projectUuid,
        params.spaceUuid,
      );

    try {
      await this.checkExistingNamesInSpace(
        [addSubspaceDto.subspaceName],
        space,
      );
      const newSubspace = this.subspaceRepository.create({
        ...addSubspaceDto,
        space,
      });

      await this.subspaceRepository.save(newSubspace);

      return new SuccessResponseDto({
        statusCode: HttpStatus.CREATED,
        data: newSubspace,
        message: 'Subspace created successfully',
      });
    } catch (error) {
      throw new HttpException(error.message, HttpStatus.INTERNAL_SERVER_ERROR);
    }
  }

  async list(
    params: GetSpaceParam,
    pageable: Partial<TypeORMCustomModelFindAllQuery>,
  ): Promise<BaseResponseDto> {
    const { communityUuid, spaceUuid, projectUuid } = params;
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      communityUuid,
      projectUuid,
      spaceUuid,
    );

    try {
      pageable.modelName = 'subspace';
      pageable.where = { space: { uuid: spaceUuid }, disabled: false };
      const customModel = TypeORMCustomModel(this.subspaceRepository);

      const { baseResponseDto, paginationResponseDto } =
        await customModel.findAll(pageable);
      return new PageResponse<SubspaceDto>(
        baseResponseDto,
        paginationResponseDto,
      );
    } catch (error) {
      throw new HttpException(error.message, HttpStatus.INTERNAL_SERVER_ERROR);
    }
  }

  async updateSubSpace(
    params: GetSubSpaceParam,
    updateSubSpaceDto: AddSubspaceDto,
  ): Promise<BaseResponseDto> {
    const { spaceUuid, communityUuid, subSpaceUuid, projectUuid } = params;
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      communityUuid,
      projectUuid,
      spaceUuid,
    );
    const subSpace = await this.subspaceRepository.findOne({
      where: { uuid: subSpaceUuid, disabled: false },
    });

    if (!subSpace) {
      throw new HttpException(
        `SubSpace with ID ${subSpaceUuid} not found`,
        HttpStatus.NOT_FOUND,
      );
    }

    try {
      Object.assign(subSpace, updateSubSpaceDto);

      await this.subspaceRepository.save(subSpace);

      return new SuccessResponseDto({
        message: `Subspace with ID ${subSpaceUuid} successfully updated`,
        data: subSpace,
        statusCode: HttpStatus.OK,
      });
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }
      throw new HttpException(
        'An error occurred while updating the subspace',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async delete(params: GetSubSpaceParam): Promise<BaseResponseDto> {
    const { spaceUuid, communityUuid, subSpaceUuid, projectUuid } = params;
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      communityUuid,
      projectUuid,
      spaceUuid,
    );

    const subspace = await this.subspaceRepository.findOne({
      where: { uuid: subSpaceUuid, disabled: false },
    });

    if (!subspace) {
      throw new HttpException(
        `Subspace with ID ${subSpaceUuid} not found`,
        HttpStatus.NOT_FOUND,
      );
    }

    try {
      await this.subspaceRepository.remove(subspace);

      return new SuccessResponseDto({
        message: `Subspace with ID ${subSpaceUuid} successfully deleted`,
        statusCode: HttpStatus.OK,
      });
    } catch (error) {
      throw new HttpException(
        'An error occurred while deleting the subspace',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  /*  async deleteSubspaces(
    deleteDtos: DeleteSubspaceDto[],
    queryRunner: QueryRunner,
  ) {
    /*   const deleteResults: { uuid: string }[] = [];

    for (const dto of deleteDtos) {
      const subspace = await this.findOne(dto.subspaceUuid);

      await queryRunner.manager.update(
        this.subspaceRepository.target,
        { uuid: dto.subspaceUuid },
        { disabled: true },
      );

      if (subspace.tags?.length) {
        const modifyTagDtos = subspace.tags.map((tag) => ({
          uuid: tag.uuid,
          action: ModifyAction.DELETE,
        }));
        await this.tagService.modifyTags(
          modifyTagDtos,
          queryRunner,
          null,
          subspace,
        );
      }

      if (subspace.devices)
        await this.deviceService.deleteSubspaceDevices(
          subspace.devices,
          queryRunner,
        );

      deleteResults.push({ uuid: dto.subspaceUuid });
    }

    return deleteResults; 
  } */

  async modifySubSpace(
    subspaceDtos: ModifySubspaceDto[],
    queryRunner: QueryRunner,
    space?: SpaceEntity,
    projectUuid?: string,
    spaceTagUpdateDtos?: ModifyTagDto[],
  ) {
    if (!subspaceDtos || subspaceDtos.length === 0) {
      return;
    }
    try {
      const addedSubspaces = [];
      const updatedSubspaces = [];

      for (const subspace of subspaceDtos) {
        switch (subspace.action) {
          case ModifyAction.ADD:
            const addedSubspace = await this.handleAddAction(
              subspace,
              space,
              queryRunner,
            );
            if (addedSubspace) addedSubspaces.push(addedSubspace);

            break;
          case ModifyAction.UPDATE:
            const updatedSubspace = await this.handleUpdateAction(
              subspace,
              queryRunner,
            );
            if (updatedSubspace) {
              updatedSubspaces.push(updatedSubspace);
            }
            break;
          case ModifyAction.DELETE:
            await this.handleDeleteAction(subspace, queryRunner);
            break;
          default:
            throw new HttpException(
              `Invalid action "${subspace.action}".`,
              HttpStatus.BAD_REQUEST,
            );
        }
      }

      const combinedSubspaces = [...addedSubspaces, ...updatedSubspaces].filter(
        (subspace) => subspace !== undefined,
      );

      if (combinedSubspaces.length > 0) {
        await this.subspaceProductAllocationService.updateSubspaceProductAllocations(
          combinedSubspaces,
          projectUuid,
          queryRunner,
          space,
          spaceTagUpdateDtos,
        );
      }
    } catch (error) {
      throw new HttpException(
        `An error occurred while modifying subspaces: ${error.message}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async unlinkModels(
    subspaces: SubspaceEntity[],
    queryRunner: QueryRunner,
  ): Promise<void> {
    if (!subspaces || subspaces.length === 0) return;

    try {
      const subspaceUuids = subspaces.map((subSpace) => subSpace.uuid);

      const allocations: SubspaceProductAllocationEntity[] = subspaces.flatMap(
        (subspace) => subspace.productAllocations || [],
      );

      if (allocations.length > 0) {
        await this.subspaceProductAllocationService.unlinkModels(
          allocations,
          queryRunner,
        );
      }

      await queryRunner.manager.update(
        SubspaceEntity,
        { uuid: In(subspaceUuids) },
        { subSpaceModel: null },
      );
    } catch (error) {
      throw new HttpException(
        `Failed to unlink subspace models: ${error instanceof Error ? error.message : 'Unknown error'}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async getOne(params: GetSubSpaceParam): Promise<BaseResponseDto> {
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      params.communityUuid,
      params.projectUuid,
      params.spaceUuid,
    );
    const subspace = await this.findOne(params.subSpaceUuid);
    return new SuccessResponseDto({
      message: `Successfully retrieved subspace`,
      data: subspace,
    });
  }

  private async handleAddAction(
    subspace: ModifySubspaceDto,
    space: SpaceEntity,
    queryRunner: QueryRunner,
  ): Promise<SubspaceEntity> {
    const createTagDtos: ProcessTagDto[] =
      subspace.tags?.map((tag) => ({
        name: tag.name as string,
        uuid: tag.tagUuid,
        productUuid: tag.productUuid as string,
      })) || [];
    const subSpace = await this.createSubspacesFromDto(
      [{ subspaceName: subspace.subspaceName, tags: createTagDtos }],
      space,
      queryRunner,
    );
    return subSpace[0];
  }

  private async handleUpdateAction(
    modifyDto: ModifySubspaceDto,
    queryRunner: QueryRunner,
  ): Promise<SubspaceEntity> {
    const subspace = await this.findOne(modifyDto.uuid);
    const updatedSubspace = await this.update(
      queryRunner,
      subspace,
      modifyDto.subspaceName,
    );
    return updatedSubspace;
  }

  async update(
    queryRunner: QueryRunner,
    subspace: SubspaceEntity,
    subspaceName?: string,
  ) {
    return await this.updateSubspaceName(queryRunner, subspace, subspaceName);
  }

  async handleDeleteAction(
    modifyDto: ModifySubspaceDto,
    queryRunner: QueryRunner,
  ): Promise<void> {
    const subspace = await this.findOne(modifyDto.uuid);

    await queryRunner.manager.update(
      this.subspaceRepository.target,
      { uuid: subspace.uuid },
      { disabled: true },
    );

    if (subspace.devices.length > 0) {
      await this.deviceService.deleteSubspaceDevices(
        subspace.devices,
        queryRunner,
      );
    }
  }

  private async findOne(subspaceUuid: string): Promise<SubspaceEntity> {
    const subspace = await this.subspaceRepository.findOne({
      where: { uuid: subspaceUuid, disabled: false },
      relations: ['tags', 'space', 'devices', 'tags.product', 'tags.device'],
    });
    if (!subspace) {
      throw new HttpException(
        `SubspaceModel with UUID ${subspaceUuid} not found.`,
        HttpStatus.NOT_FOUND,
      );
    }
    return subspace;
  }

  async updateSubspaceName(
    queryRunner: QueryRunner,
    subSpace: SubspaceEntity,
    subspaceName?: string,
  ): Promise<SubspaceEntity> {
    if (subspaceName) {
      subSpace.subspaceName = subspaceName;
      return await queryRunner.manager.save(subSpace);
    }
    return subSpace;
  }

  private async checkForDuplicateNames(names: string[]): Promise<void> {
    const seenNames = new Set<string>();
    const duplicateNames = new Set<string>();

    for (const name of names) {
      if (!seenNames.add(name)) {
        duplicateNames.add(name);
      }
    }

    if (duplicateNames.size > 0) {
      throw new HttpException(
        `Duplicate subspace names found: ${[...duplicateNames].join(', ')}`,
        HttpStatus.CONFLICT,
      );
    }
  }

  private async checkExistingNamesInSpace(
    names: string[],
    space: SpaceEntity,
  ): Promise<void> {
    const existingNames = await this.subspaceRepository.find({
      select: ['subspaceName'],
      where: {
        subspaceName: In(names),
        space: {
          uuid: space.uuid,
        },
        disabled: false,
      },
    });

    if (existingNames.length > 0) {
      const existingNamesList = existingNames
        .map((e) => e.subspaceName)
        .join(', ');
      throw new HttpException(
        `Subspace names already exist in the space: ${existingNamesList}`,
        HttpStatus.BAD_REQUEST,
      );
    }
  }

  private async validateName(
    names: string[],
    space: SpaceEntity,
  ): Promise<void> {
    await this.checkForDuplicateNames(names);
    await this.checkExistingNamesInSpace(names, space);
  }
  extractTagsFromSubspace(addSubspaceDto: AddSubspaceDto[]): ProcessTagDto[] {
    return addSubspaceDto.flatMap((subspace) => subspace.tags || []);
  }
  async clearSubspaces(subspaceUuids: string[], queryRunner: QueryRunner) {
    try {
      await queryRunner.manager.update(
        SubspaceEntity,
        { uuid: In(subspaceUuids) },
        { disabled: true },
      );

      await this.subspaceProductAllocationService.clearAllAllocations(
        subspaceUuids,
        queryRunner,
      );
    } catch (error) {
      throw new HttpException(
        error instanceof HttpException
          ? error.message
          : 'An unexpected error occurred while clearing subspaces',
        error instanceof HttpException
          ? error.getStatus()
          : HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }
}
