import { ModifyAction } from '@app/common/constants/modify-action.enum';
import { TagRepository } from '@app/common/modules/space';
import { TagModel } from '@app/common/modules/space-model';
import { SpaceEntity } from '@app/common/modules/space/entities/space.entity';
import { SubspaceEntity } from '@app/common/modules/space/entities/subspace/subspace.entity';
import { TagEntity } from '@app/common/modules/space/entities/tag.entity';
import { HttpException, HttpStatus, Injectable } from '@nestjs/common';
import { ProductService } from 'src/product/services';
import { ModifySubspaceDto } from 'src/space/dtos';
import { ModifyTagDto } from 'src/space/dtos/tag/modify-tag.dto';
import { ProcessTagDto } from 'src/tags/dtos';
import { QueryRunner } from 'typeorm';

@Injectable()
export class TagService {
  constructor(
    private readonly tagRepository: TagRepository,
    private readonly productService: ProductService,
  ) {}

  async createTags(
    tags: ProcessTagDto[],
    queryRunner: QueryRunner,
    space?: SpaceEntity,
    subspace?: SubspaceEntity,
    additionalTags?: ProcessTagDto[],
    tagsToDelete?: ModifyTagDto[],
  ): Promise<TagEntity[]> {
    this.validateTagsInput(tags);

    const combinedTags = this.combineTags(tags, additionalTags);
    this.ensureNoDuplicateTags(combinedTags);

    const tagEntitiesToCreate = tags.filter((tagDto) => !tagDto.uuid);
    const tagEntitiesToUpdate = tags.filter((tagDto) => !!tagDto.uuid);

    try {
      const createdTags = await this.bulkSaveTags(
        tagEntitiesToCreate,
        queryRunner,
        space,
        subspace,
        tagsToDelete,
      );
      const updatedTags = await this.moveTags(
        tagEntitiesToUpdate,
        queryRunner,
        space,
        subspace,
      );

      return [...createdTags, ...updatedTags];
    } catch (error) {
      throw this.handleUnexpectedError('Failed to save tags', error);
    }
  }

  async bulkSaveTags(
    tags: ProcessTagDto[],
    queryRunner: QueryRunner,
    space?: SpaceEntity,
    subspace?: SubspaceEntity,
    tagsToDelete?: ModifyTagDto[],
  ): Promise<TagEntity[]> {
    if (!tags.length) {
      return [];
    }

    const tagEntities = await Promise.all(
      tags.map((tagDto) =>
        this.prepareTagEntity(
          tagDto,
          queryRunner,
          space,
          subspace,
          tagsToDelete,
        ),
      ),
    );

    try {
      return await queryRunner.manager.save(tagEntities);
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }

      throw new HttpException(
        `Failed to save tag models due to an unexpected error: ${error.message}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async moveTags(
    tags: ProcessTagDto[],
    queryRunner: QueryRunner,
    space?: SpaceEntity,
    subspace?: SubspaceEntity,
  ): Promise<TagEntity[]> {
    if (!tags.length) {
      return [];
    }

    try {
      return await Promise.all(
        tags.map(async (tagDto) => {
          try {
            const tag = await this.getTagByUuid(tagDto.uuid);
            if (!tag) {
              throw new HttpException(
                `Tag with UUID ${tagDto.uuid} not found.`,
                HttpStatus.NOT_FOUND,
              );
            }

            if (subspace && subspace.space) {
              await queryRunner.manager.update(
                this.tagRepository.target,
                { uuid: tag.uuid },
                { subspace },
              );
              tag.subspace = subspace;
            }

            if (!subspace && space) {
              await queryRunner.manager.update(
                this.tagRepository.target,
                { uuid: tag.uuid },
                { subspace: null },
              );
              tag.subspace = null;
            }

            return tag;
          } catch (error) {
            throw error;
          }
        }),
      );
    } catch (error) {
      throw new HttpException(
        `Failed to move tags due to an unexpected error: ${error.message}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async createTagsFromModel(
    queryRunner: QueryRunner,
    tagModels: TagModel[],
    space?: SpaceEntity,
    subspace?: SubspaceEntity,
  ): Promise<void> {
    if (!tagModels?.length) return;

    const tags = tagModels.map((model) =>
      queryRunner.manager.create(this.tagRepository.target, {
        tag: model.tag,
        space: space || undefined,
        subspace: subspace || undefined,
        product: model.product,
      }),
    );

    await queryRunner.manager.save(tags);
  }

  async updateTag(
    tag: ModifyTagDto,
    queryRunner: QueryRunner,
    space?: SpaceEntity,
    subspace?: SubspaceEntity,
  ): Promise<TagEntity> {
    try {
      const existingTag = await this.getTagByUuid(tag.tagUuid);

      const contextSpace = space ?? subspace?.space;

      if (contextSpace && tag.name !== existingTag.tag) {
        await this.checkTagReuse(
          tag.name,
          existingTag.product.uuid,
          contextSpace,
        );
      }

      return await queryRunner.manager.save(
        Object.assign(existingTag, { tag: tag.name }),
      );
    } catch (error) {
      throw this.handleUnexpectedError('Failed to update tags', error);
    }
  }

  async updateTagsFromModel(
    model: TagModel,
    queryRunner: QueryRunner,
  ): Promise<void> {
    try {
      const tags = await this.tagRepository.find({
        where: {
          model: {
            uuid: model.uuid,
          },
        },
      });

      if (!tags.length) return;

      await queryRunner.manager.update(
        this.tagRepository.target,
        { model: { uuid: model.uuid } },
        { tag: model.tag },
      );
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }

      throw new HttpException(
        `Failed to update tags for model with UUID: ${model.uuid}. Reason: ${error.message}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async deleteTags(tagUuids: string[], queryRunner: QueryRunner) {
    if (!tagUuids?.length) return;

    try {
      await Promise.all(
        tagUuids.map((id) =>
          queryRunner.manager.update(
            this.tagRepository.target,
            { uuid: id },
            { disabled: true, device: null },
          ),
        ),
      );
      return { message: 'Tags deleted successfully', tagUuids };
    } catch (error) {
      throw this.handleUnexpectedError('Failed to delete tags', error);
    }
  }

  async deleteTagFromModel(modelUuid: string, queryRunner: QueryRunner) {
    try {
      const tags = await this.tagRepository.find({
        where: {
          model: {
            uuid: modelUuid,
          },
        },
      });

      if (!tags.length) return;

      await queryRunner.manager.update(
        this.tagRepository.target,
        { model: { uuid: modelUuid } },
        { disabled: true, device: null },
      );
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }

      throw new HttpException(
        `Failed to update tags for model with UUID: ${modelUuid}. Reason: ${error.message}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async modifyTags(
    tags: ModifyTagDto[],
    queryRunner: QueryRunner,
    space?: SpaceEntity,
    subspace?: SubspaceEntity,
  ): Promise<void> {
    if (!tags?.length) return;

    try {
      const tagsToDelete = tags.filter(
        (tag) => tag.action === ModifyAction.DELETE,
      );

      await Promise.all(
        tags.map(async (tag) => {
          switch (tag.action) {
            case ModifyAction.ADD:
              await this.createTags(
                [
                  {
                    name: tag.name,
                    productUuid: tag.productUuid,
                    uuid: tag.tagUuid,
                  },
                ],
                queryRunner,
                space,
                subspace,
                null,
                tagsToDelete,
              );
              break;
            case ModifyAction.UPDATE:
              await this.updateTag(tag, queryRunner, space, subspace);
              break;
            case ModifyAction.DELETE:
              await this.deleteTags([tag.tagUuid], queryRunner);
              break;
            default:
              throw new HttpException(
                `Invalid action "${tag.action}" provided.`,
                HttpStatus.BAD_REQUEST,
              );
          }
        }),
      );
    } catch (error) {
      throw this.handleUnexpectedError('Failed to modify tags', error);
    }
  }

  async unlinkModels(tags: TagEntity[], queryRunner: QueryRunner) {
    if (!tags?.length) return;

    try {
      tags.forEach((tag) => {
        tag.model = null;
      });

      await queryRunner.manager.save(tags);
    } catch (error) {
      throw this.handleUnexpectedError('Failed to unlink tag models', error);
    }
  }

  private findDuplicateTags(tags: ProcessTagDto[]): string[] {
    const seen = new Map<string, boolean>();
    const duplicates: string[] = [];

    tags.forEach((tagDto) => {
      const key = `${tagDto.productUuid}-${tagDto.name}`;
      if (seen.has(key)) {
        duplicates.push(`${tagDto.name} for Product: ${tagDto.productUuid}`);
      } else {
        seen.set(key, true);
      }
    });

    return duplicates;
  }

  private async checkTagReuse(
    tag: string,
    productUuid: string,
    space: SpaceEntity,
    tagsToDelete?: ModifyTagDto[],
  ): Promise<void> {
    const { uuid: spaceUuid } = space;

    const tagExists = await this.tagRepository.find({
      where: [
        {
          tag,
          product: { uuid: productUuid },
          disabled: false,
        },
        {
          tag,
          subspace: { space: { uuid: spaceUuid } },
          product: { uuid: productUuid },
          disabled: false,
        },
      ],
    });

    const filteredTagExists = tagExists.filter(
      (existingTag) =>
        !tagsToDelete?.some(
          (deleteTag) => deleteTag.tagUuid === existingTag.uuid,
        ),
    );

    if (filteredTagExists.length > 0) {
      throw new HttpException(`Tag can't be reused`, HttpStatus.CONFLICT);
    }
  }

  private async prepareTagEntity(
    tagDto: ProcessTagDto,
    queryRunner: QueryRunner,
    space?: SpaceEntity,
    subspace?: SubspaceEntity,
    tagsToDelete?: ModifyTagDto[],
  ): Promise<TagEntity> {
    try {
      const product = await this.productService.findOne(tagDto.productUuid);

      if (!product) {
        throw new HttpException(
          `Product with UUID ${tagDto.productUuid} not found.`,
          HttpStatus.NOT_FOUND,
        );
      }

      if (space) {
        await this.checkTagReuse(
          tagDto.name,
          tagDto.productUuid,
          space,
          tagsToDelete,
        );
      } else if (subspace && subspace.space) {
        await this.checkTagReuse(
          tagDto.name,
          tagDto.productUuid,
          subspace.space,
        );
      } else {
        throw new HttpException(
          `Invalid subspace or space provided.`,
          HttpStatus.BAD_REQUEST,
        );
      }

      return queryRunner.manager.create(TagEntity, {
        tag: tagDto.name,
        product: product.data,
        space: space,
        subspace: subspace,
      });
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }
      throw new HttpException(
        `An error occurred while preparing the tag entity: ${error.message}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  private async getTagByUuid(uuid: string): Promise<TagEntity> {
    const tag = await this.tagRepository.findOne({
      where: { uuid },
      relations: ['product'],
    });

    if (!tag) {
      throw new HttpException(
        `Tag with ID ${uuid} not found.`,
        HttpStatus.NOT_FOUND,
      );
    }
    return tag;
  }

  private handleUnexpectedError(
    message: string,
    error: unknown,
  ): HttpException {
    if (error instanceof HttpException) throw error;
    return new HttpException(
      `${message}: ${(error as Error)?.message || 'Unknown error'}`,
      HttpStatus.INTERNAL_SERVER_ERROR,
    );
  }

  private combineTags(
    primaryTags: ProcessTagDto[],
    additionalTags?: ProcessTagDto[],
  ): ProcessTagDto[] {
    return additionalTags ? [...primaryTags, ...additionalTags] : primaryTags;
  }

  private ensureNoDuplicateTags(tags: ProcessTagDto[]): void {
    const duplicates = this.findDuplicateTags(tags);

    if (duplicates.length > 0) {
      throw new HttpException(
        `Duplicate tags found: ${duplicates.join(', ')}`,
        HttpStatus.BAD_REQUEST,
      );
    }
  }

  private validateTagsInput(tags: ProcessTagDto[]): void {
    if (!tags?.length) {
      return;
    }
  }

  getSubspaceTagsToBeAdded(
    spaceTags?: ModifyTagDto[],
    subspaceModels?: ModifySubspaceDto[],
  ): ModifyTagDto[] {
    if (!subspaceModels || subspaceModels.length === 0) {
      return spaceTags;
    }

    const spaceTagsToDelete = spaceTags?.filter(
      (tag) => tag.action === 'delete',
    );

    const tagsToAdd = subspaceModels.flatMap(
      (subspace) => subspace.tags?.filter((tag) => tag.action === 'add') || [],
    );

    const commonTagUuids = new Set(
      tagsToAdd
        .filter((tagToAdd) =>
          spaceTagsToDelete.some(
            (tagToDelete) => tagToAdd.tagUuid === tagToDelete.tagUuid,
          ),
        )
        .map((tag) => tag.tagUuid),
    );

    const remainingTags = spaceTags.filter(
      (tag) => !commonTagUuids.has(tag.tagUuid), // Exclude tags in commonTagUuids
    );

    return remainingTags;
  }

  getModifiedSubspaces(
    spaceTags?: ModifyTagDto[],
    subspaceModels?: ModifySubspaceDto[],
  ): ModifySubspaceDto[] {
    if (!subspaceModels || subspaceModels.length === 0) {
      return [];
    }

    // Extract tags marked for addition in spaceTags
    const spaceTagsToAdd = spaceTags?.filter((tag) => tag.action === 'add');

    const subspaceTagsToAdd = subspaceModels.flatMap(
      (subspace) => subspace.tags?.filter((tag) => tag.action === 'add') || [],
    );

    const subspaceTagsToDelete = subspaceModels.flatMap(
      (subspace) =>
        subspace.tags?.filter((tag) => tag.action === 'delete') || [],
    );

    const subspaceTagsToDeleteUuids = new Set(
      subspaceTagsToDelete.map((tag) => tag.tagUuid),
    );

    const commonTagsInSubspaces = subspaceTagsToAdd.filter((tag) =>
      subspaceTagsToDeleteUuids.has(tag.tagUuid),
    );

    // Find UUIDs of tags that are common between spaceTagsToAdd and subspace tags marked for deletion
    const commonTagUuids = new Set(
      spaceTagsToAdd
        .flatMap((tagToAdd) =>
          subspaceModels.flatMap(
            (subspace) =>
              subspace.tags?.filter(
                (tagToDelete) =>
                  tagToDelete.action === 'delete' &&
                  tagToAdd.tagUuid === tagToDelete.tagUuid,
              ) || [],
          ),
        )
        .map((tag) => tag.tagUuid),
    );

    // Modify subspaceModels by removing tags with UUIDs present in commonTagUuids
    let modifiedSubspaces = subspaceModels.map((subspace) => ({
      ...subspace,
      tags:
        subspace.tags?.filter((tag) => !commonTagUuids.has(tag.tagUuid)) || [],
    }));

    modifiedSubspaces = modifiedSubspaces.map((subspace) => ({
      ...subspace,
      tags:
        subspace.tags?.filter(
          (tag) =>
            !(
              tag.action === 'delete' &&
              commonTagsInSubspaces.some(
                (commonTag) => commonTag.tagUuid === tag.tagUuid,
              )
            ),
        ) || [],
    }));

    return modifiedSubspaces;
  }
}
